chapter \<open>Trace Outer_Syntax.command down to storing proofs\<close>

theory Downto_Synchronized
  imports Pure
begin
text \<open>
  We assume, that proofs are stored in a Synchronized.*.
  We want to re-use as much of Isabelle's proof machinery.
\<close>

section \<open>Tools for tracing\<close>
ML \<open>(*implementation.pdf p.15: output like a record..*)
val x = 42;
val y = true;
\<close> ML \<open>
writeln (@{make_string} {x = x, y = y}); (*
{x = 42, y = true}                                                  
val it = (): unit*)
\<close> ML \<open>
@{print} {x = x, y = y};                 (* 1st line normal Output:
{x = 42, y = true}\<^here> 
val it = {x = 42, y = true}: {x: int, y: bool}*)
\<close> ML \<open>
@{print\<open>tracing\<close>} {x = x, y = y};        (* 1st line blue as tracing:
{x = 42, y = true}\<^here> 
val it = {x = 42, y = true}: {x: int, y: bool}*)
\<close>

section \<open>Trace code down to Table for..\<close>
subsection \<open>Outer_Syntax.command..spark_vc :: thy_goal\<close>
text \<open>
  spark_vc starts an interactive proof, given data for initialising the proof.
  conclusions:
* spark_vc <--> ISAC: Calculation \<open>~~/Examples/exp_Statics_Biegel_Timischl_7-70\<close>
* for how to open the data-file \<open>exp_Statics_Biegel_Timischl_7-70\<close> see spark_open
\<close> ML \<open>
\<close>
declare [[ML_print_depth = 999]]
ML \<open>
signature SPECIFICATION =
sig
  val theorem: bool -> string -> Method.text option ->
    (thm list list -> local_theory -> local_theory) ->
      Attrib.binding -> string list -> Element.context_i list -> Element.statement_i ->
        bool -> local_theory ->
          Proof.state
end
(**)structure Specification: SPECIFICATION =
struct(**)
fun gen_theorem (_: bool)
  (_: 'a -> local_theory -> Proof.context)
    (_: local_theory -> Token.src -> Token.src)
      (_: ('b, 'c, 'd) Element.ctxt list -> ('e, 'f) Element.stmt -> Proof.context ->
          ((binding * Token.src list) * (term * term list) list) list * Proof.context)
        (_: bool)
          (_: string)
            (_: Method.text option)
              (_: thm list list -> local_theory -> Proof.context)
                (_: binding * Token.src list)
                  (_: 'a)
                    (_: ('b, 'c, 'd) Element.ctxt list)
                      (_: ('e, 'f) Element.stmt)
                        (_: bool)
                          (_: local_theory) =
  Proof.init @{context};

val theorem = (*<----- orig*)
  gen_theorem false Bundle.includes (K I) Expression.cert_statement;
(*
   Signature: val theorem: bool -> string -> Method.text option -> (thm list list -> local_theory -> local_theory) -> Attrib.binding -> string list -> ...
   Structure: val theorem: Proof.state -> Proof.state
*)
(**)end(**)
\<close> ML \<open>

signature EXPRESSION =
sig
val cert_statement: Element.context_i list -> Element.statement_i ->
    Proof.context -> (Attrib.binding * (term * term list) list) list * Proof.context
(** )val read_statement: Element.context list -> Element.statement ->
    Proof.context -> (Attrib.binding * (term * term list) list) list * Proof.context( **)
end
(**)structure Expression : EXPRESSION =
struct(**)
fun prep_statement (*dummy*)
(* 1 *)(_: {do_close: bool, fixed_frees: bool, strict: bool} ->
(*2-4*)  'a list * 'b list ->
(* 5 *)    ('c -> 'c) ->
(* 6 *)      'd ->
(* 7 *)        'e ->
(* 8 *)          Proof.context ->
(*9-f*)      ('f * 'g * 'h * 'i list * 'j) *
(* g *)      'k
(*   *))
(*   *)  (_: 'i -> Proof.context -> 'l * Proof.context)
(*   *)    (_: 'd) (_: 'e) (_: Proof.context) =
(* e *)      ( []: (Attrib.binding * (term * term list) list) list,
(*   *)        @{context}
(*   *)      )
datatype ctxt = datatype Element.ctxt;
datatype 'term map = Positional of 'term option list | Named of (string * 'term) list
fun cert_full_context_statement (*dummy*)
(* 1 *)(_: {do_close: 'a, fixed_frees: bool, strict: bool})
(* 2 *)  (_: (string * ((string * bool) * (term map * (('b * 'c list) * term) list))) list *
(* 3 *)    (binding * typ option * mixfix) list
(* 4 *)  )
(* 5 *)    (_: Proof.context -> Proof.context)
(* 6 *)      (_: (typ, term, 'd) ctxt list)
(* 7 *)        (_: (typ, term) Element.stmt)
(* 8 *)          (_: Proof.context) =
(* 9 *)            ( (
(* a *)                []:((bstring * typ) * mixfix) list,
(* b *)                []:(string * morphism) list,
(* c *)                []:(('b * 'c list) * term) list list,
(* d *)                []:(typ, 'e, 'f) ctxt list,
(* e *)                []:(Attrib.binding * (term * term list) list) list
(* f *)              ),
(* g *)              ([]:(string * typ) list, @{context}) 
(* h *)            );
(*Element.activate_i: Element.context_i -> Proof.context -> Element.context_i * Proof.context;*)

fun cert_statement x = prep_statement cert_full_context_statement Element.activate_i x;
(** )fun read_statement x = prep_statement read_full_context_statement Element.activate x;( **)
(**)end(**)
\<close> ML \<open>

\<close> ML \<open>

signature ELEMENT =
sig
  datatype ('typ, 'term, 'fact) ctxt =
    Fixes of (binding * 'typ option * mixfix) list |
    Constrains of (string * 'typ) list |
    Assumes of (Attrib.binding * ('term * 'term list) list) list |
    Defines of (Attrib.binding * ('term * 'term list)) list |
    Notes of string * (Attrib.binding * ('fact * Token.src list) list) list |
    Lazy_Notes of string * (binding * 'fact lazy)
  type context_i
  val activate_i: context_i -> Proof.context -> context_i * Proof.context
(*val activate: (typ, term, Facts.ref) ctxt -> Proof.context -> context_i * Proof.context*)
end
\<close> ML \<open>
structure Element: ELEMENT =
struct
datatype ('typ, 'term, 'fact) ctxt =
  Fixes of (binding * 'typ option * mixfix) list |
  Constrains of (string * 'typ) list |
  Assumes of (Attrib.binding * ('term * 'term list) list) list |
  Defines of (Attrib.binding * ('term * 'term list)) list |
  Notes of string * (Attrib.binding * ('fact * Token.src list) list) list |
  Lazy_Notes of string * (binding * 'fact lazy)
type context_i = (typ, term, thm list) ctxt;
fun map_ctxt_attrib (_: Token.src -> Token.src) (xxx : ('typ, 'term, 'fact) ctxt) = xxx
datatype generic = Theory of theory | Proof of Proof.context
fun init (_: context_i) (xxx: Context.generic) = xxx

fun activate_i elem ctxt =
  let
    val elem' =
      (case (map_ctxt_attrib o map) Token.init_assignable elem of
        Defines defs =>
          Defines (defs |> map (fn ((a, atts), (t, ps)) =>
            ((Thm.def_binding_optional
              (Binding.name (#1 (#1 (Local_Defs.cert_def ctxt (K []) t)))) a, atts),
              (t, ps))))
      | e => e);
    val ctxt' = Context.proof_map (init elem') ctxt;
  in ((map_ctxt_attrib o map) Token.closure elem', ctxt') end;
end


signature PRIVATE_CONTEXT =
sig
  type generic
  val proof_map: (generic -> generic) -> Proof.context -> Proof.context
end
structure Context: PRIVATE_CONTEXT =
struct
datatype generic = Theory of theory | Proof of Proof.context;
fun the_proof (_: generic) = @{context} (*dummy*)

fun proof_map f = the_proof o f o Proof;

(*/--------------------? related ?-------------------------------------------\* )
fun declare_theory_data pos empty extend merge =
  let
    val k = serial ();
    val kind = {pos = pos, empty = empty, extend = extend, merge = merge};
    val _ = Synchronized.change kinds (Datatab.update (k, kind));
  in k end;                          (*^^^^^^^^^^^^^^*)

val extend_data = Datatab.map invoke_extend;
fun merge_data thys = Datatab.join (invoke_merge thys) o apply2 extend_data;
                   (*^^^^^^^^^^^^^^*)
(*private copy avoids potential conflict of table exceptions*)
structure Datatab = Table(type key = int val ord = int_ord);
                   (*^^^^*)
( *\--------------------? related ?-------------------------------------------/*)
end

\<close> ML \<open>
signature CHANGE_TABLE =
sig
  structure Table: TABLE
end
signature TABLE =
sig
  type key
  type 'a table
  val update: key * 'a -> 'a table -> 'a table
  val join: (key -> 'a * 'a -> 'a) (*exception SAME*) ->
    'a table * 'a table -> 'a table                                    (*exception DUP*)
end
(**)functor Table(Key: KEY): TABLE =
struct
type key = Key.key;(** )
type key = int val ord = int_ord;( **)
datatype 'a table =
  Empty |
  Branch2 of 'a table * (key * 'a) * 'a table |
  Branch3 of 'a table * (key * 'a) * 'a table * (key * 'a) * 'a table;
fun is_empty (_: 'a table) = true (*dummy*)
fun modify (_: key) (_: 'b -> 'c) (xxx : 'a table) = xxx (*dummy*)

fun update (key, x) tab = modify key (fn _ => x) tab;

(*/------------------------------- construct dummy above -------------------------\* )
(* find sig.of modify via arg..*)
fun update modify (key, x) tab = modify key (fn _ => x) tab; 
       update: ('a -> ('b -> 'c) -> 'd -> 'e) -> 'a  * 'c -> 'd       -> 'e
(*sig: update:                                   key * 'a -> 'a table -> 'a table*)
( *\------------------------------- construct dummy above -------------------------/*)

fun fold_table (*dummy*)
  (_: key * 'a -> 'b table -> 'b table) (_: 'c table) (xxx: 'c table) = xxx
(* simultaneous modifications *)
fun join f (table1, table2) =
  let
    fun add (key, y) tab = modify key (fn NONE => y | SOME x => f key (x, y)) tab;
  in
    if pointer_eq (table1, table2) then table1
    else if is_empty table1 then table2
    else fold_table add table2 table1
  end;
(*/------------------------------- construct dummy above -------------------------\* )
(* find sig.of modify via arg..*)
fun join fold_table f (table1, table2) =
  let
    fun add (key, y) tab = modify key (fn NONE => y | SOME x => f key (x, y)) tab;
  in
    if pointer_eq (table1, table2) then table1
    else if is_empty table1 then table2
    else fold_table add table2 table1
  end;

       join: ((key * 'a -> 'b table -> 'b table) -> 'c table -> 'c table -> 'c table) ->
               (key -> 'd * 'a -> 'a) -> 'c table * 'c table -> 'c table
(*sig: join:   (key -> 'a * 'a -> 'a) -> 'a table * 'a table -> 'a table*)
( *\------------------------------- construct dummy above -------------------------/*)

(**)end(**)
\<close> ML \<open>
\<close> ML \<open>
\<close>

subsection \<open>lemma (from Pure.thy)\<close>
theorem meta_mp:
  assumes "PROP P \<Longrightarrow> PROP Q" and "PROP P"
  shows "PROP Q"
    by (rule \<open>PROP P \<Longrightarrow> PROP Q\<close> [OF \<open>PROP P\<close>])




end